@@ -87,7 +87,7 @@ class DimensionalityError(ValueError):
     """Raised when trying to convert between incompatible units.
     """
 
-    def __init__(self, units1, units2, dim1=None, dim2=None, extra_msg=None):
+    def __init__(self, units1, units2, dim1=None, dim2=None, extra_msg=''):
         super(DimensionalityError, self).__init__()
         self.units1 = units1
         self.units2 = units2
@@ -103,12 +103,25 @@ class DimensionalityError(ValueError):
             dim1 = ''
             dim2 = ''
 
-        msg = "Cannot convert from '{0}'{1} to '{2}'{3}".format(self.units1, dim1, self.units2, dim2)
+        msg = "Cannot convert from '{0}'{1} to '{2}'{3}" + self.extra_msg
 
-        if self.extra_msg:
-            return msg + self.extra_msg
+        return msg.format(self.units1, dim1, self.units2, dim2)
 
-        return msg
+
+class OffsetUnitCalculusError(ValueError):
+    """Raised on ambiguous operations with offset units.
+    """
+    def __init__(self, units1, units2='', extra_msg=''):
+        super(ValueError, self).__init__()
+        self.units1 = units1
+        self.units2 = units2
+        self.extra_msg = extra_msg
+
+    def __str__(self):
+        msg = ("Ambiguous operation with offset unit (%s)." %
+               ', '.join(['%s' % u for u in [self.units1, self.units2] if u])
+               + self.extra_msg)
+        return msg.format(self.units1, self.units2)
 
 
 class Converter(object):
@@ -244,7 +257,7 @@ class Definition(object):
 
 
 def _is_dim(name):
-    return name.startswith('[') and name.endswith(']')
+    return name[0] == '[' and name[-1] == ']'
 
 
 class PrefixDefinition(Definition):
@@ -289,7 +302,7 @@ class UnitDefinition(Definition):
                                  'Base units must be referenced only to dimensions. '
                                  'Derived units must be referenced only to units.')
             self.reference = UnitsContainer(converter.items())
-            if 'offset' in modifiers:
+            if modifiers.get('offset', 0.) != 0.:
                 converter = OffsetConverter(converter.scale, modifiers['offset'])
             else:
                 converter = ScaleConverter(converter.scale)
@@ -445,14 +458,19 @@ class UnitRegistry(object):
                      Empty to load the default definition file.
                      None to leave the UnitRegistry empty.
     :param force_ndarray: convert any input, scalar or not to a numpy.ndarray.
-    :param default_to_delta: In the context of a multiplication of units, interpret
+    :param default_as_delta: In the context of a multiplication of units, interpret
                              non-multiplicative units as their *delta* counterparts.
+    :autoconvert_offset_to_baseunit: If True converts offset units in quantites are
+                                     converted to their base units in multiplicative
+                                     context. If False no conversion happens.
     :param on_redefinition: action to take in case a unit is redefined.
                             'warn', 'raise', 'ignore'
     :type on_redefintion: str
     """
 
-    def __init__(self, filename='', force_ndarray=False, default_to_delta=True, on_redefinition='warn'):
+    def __init__(self, filename='', force_ndarray=False, default_as_delta=True,
+                 autoconvert_offset_to_baseunit=False,
+                 on_redefinition='warn'):
         self.Quantity = build_quantity_class(self, force_ndarray)
         self.Measurement = build_measurement_class(self, force_ndarray)
 
@@ -491,9 +509,16 @@ class UnitRegistry(object):
         #: Maps dimensionality (_freeze(UnitsContainer)) to Units (_freeze(UnitsContainer))
         self._dimensionality_cache = TransformDict(_freeze)
 
+        #: Cache the unit name associated to user input. ('mV' -> 'millivolt')
+        self._parse_unit_cache = dict()
+
         #: When performing a multiplication of units, interpret
         #: non-multiplicative units as their *delta* counterparts.
-        self.default_to_delta = default_to_delta
+        self.default_as_delta = default_as_delta
+
+        # Determines if quantities with offset units are converted to their
+        # base units on multiplication and division.
+        self.autoconvert_offset_to_baseunit = autoconvert_offset_to_baseunit
 
         if filename == '':
             self.load_definitions('default_en.txt', True)
@@ -504,6 +529,9 @@ class UnitRegistry(object):
 
         self._build_cache()
 
+    def __name__(self):
+        return 'UnitRegistry'
+
     def __getattr__(self, item):
         return self.Quantity(1, item)
 
@@ -697,7 +725,8 @@ class UnitRegistry(object):
 
             _adder(alias, definition)
 
-        if isinstance(definition.converter, OffsetConverter):
+        # define additional "delta_" units for units with an offset
+        if getattr(definition.converter, "offset", 0.0) != 0.0:
             d_name = 'delta_' + definition.name
             if definition.symbol:
                 d_symbol = 'Δ' + definition.symbol
@@ -710,7 +739,7 @@ class UnitRegistry(object):
                     return '[delta_' + _name[1:]
                 return 'delta_' + _name
 
-            d_reference = UnitsContainer(dict((prep(ref), value)
+            d_reference = UnitsContainer(dict((ref, value)
                                          for ref, value in definition.reference.items()))
 
             self.define(UnitDefinition(d_name, d_symbol, d_aliases,
@@ -818,7 +847,7 @@ class UnitRegistry(object):
             return ''
 
         try:
-            return self._units[name_or_alias].name
+            return self._units[name_or_alias]._name
         except KeyError:
             pass
 
@@ -857,15 +886,17 @@ class UnitRegistry(object):
 
         return self._prefixes[prefix].symbol + self._units[unit_name].symbol
 
+    def _get_symbol(self, name):
+        return self._units[name].symbol
+
     def get_dimensionality(self, input_units):
         """Convert unit or dict of units or dimensions to a dict of base dimensions
 
         :param input_units:
         :return: dimensionality
         """
-        dims = UnitsContainer()
         if not input_units:
-            return dims
+            return UnitsContainer()
 
         if isinstance(input_units, string_types):
             input_units = ParserHelper.from_string(input_units)
@@ -873,25 +904,32 @@ class UnitRegistry(object):
         if input_units in self._dimensionality_cache:
             return copy.copy(self._dimensionality_cache[input_units])
 
-        for key, value in input_units.items():
-            if _is_dim(key):
-                reg = self._dimensions[key]
-                if reg.is_base:
-                    dims.add(key, value)
-                else:
-                    dims *= self.get_dimensionality(reg.reference) ** value
-            else:
-                reg = self._units[self.get_name(key)]
-                if reg.is_base:
-                    dims *= reg.reference ** value
-                else:
-                    dims *= self.get_dimensionality(reg.reference) ** value
+        accumulator = defaultdict(float)
+        self._get_dimensionality_recurse(input_units, 1.0, accumulator)
+
+        dims = UnitsContainer(dict((k, v) for k, v in accumulator.items() if v != 0.))
 
         if '[]' in dims:
             del dims['[]']
 
+        self._dimensionality_cache[input_units] = copy.copy(dims)
+
         return dims
 
+    def _get_dimensionality_recurse(self, ref, exp, accumulator):
+        for key in ref:
+            exp2 = exp*ref[key]
+            if _is_dim(key):
+                reg = self._dimensions[key]
+                if reg.is_base:
+                    accumulator[key] += exp2
+                elif reg.reference is not None:
+                    self._get_dimensionality_recurse(reg.reference, exp2, accumulator)
+            else:
+                reg = self._units[self.get_name(key)]
+                if reg.reference is not None:
+                    self._get_dimensionality_recurse(reg.reference, exp2, accumulator)
+
     def get_base_units(self, input_units, check_nonmult=True):
         """Convert unit or dict of units to the base units.
 
@@ -914,27 +952,32 @@ class UnitRegistry(object):
         if check_nonmult and input_units in self._base_units_cache:
             return copy.deepcopy(self._base_units_cache[input_units])
 
-        factor = 1.
-        units = UnitsContainer()
-        for key, value in input_units.items():
-            key = self.get_name(key)
-            reg = self._units[key]
-            if reg.is_base:
-                units.add(key, value)
-            else:
-                fac, uni = self.get_base_units(reg.reference, check_nonmult=False)
-                if factor is not None:
-                    factor *= (reg.converter.scale * fac) ** value
-                units *= uni ** value
+        accumulators = [1., defaultdict(float)]
+        self._get_base_units(input_units, 1.0, accumulators)
+
+        factor = accumulators[0]
+        units = UnitsContainer(dict((k, v) for k, v in accumulators[1].items() if v != 0.))
 
         # Check if any of the final units is non multiplicative and return None instead.
         if check_nonmult:
             for unit in units.keys():
-                if not isinstance(self._units[unit].converter, ScaleConverter):
+                if not self._units[unit].converter.is_multiplicative:
                     return None, units
 
         return factor, units
 
+    def _get_base_units(self, ref, exp, accumulators):
+        for key in ref:
+            exp2 = exp*ref[key]
+            key = self.get_name(key)
+            reg = self._units[key]
+            if reg.is_base:
+                accumulators[1][key] += exp2
+            else:
+                accumulators[0] *= reg._converter.scale ** exp2
+                if reg.reference is not None:
+                    self._get_base_units(reg.reference, exp2, accumulators)
+
     def get_compatible_units(self, input_units):
         if not input_units:
             return 1., UnitsContainer()
@@ -967,9 +1010,9 @@ class UnitRegistry(object):
         :return: converted value
         """
         if isinstance(src, string_types):
-            src = ParserHelper.from_string(src)
+            src = self.parse_units(src)
         if isinstance(dst, string_types):
-            dst = ParserHelper.from_string(dst)
+            dst = self.parse_units(dst)
         if src == dst:
             return value
 
@@ -997,36 +1040,59 @@ class UnitRegistry(object):
         if src_dim != dst_dim:
             raise DimensionalityError(src, dst, src_dim, dst_dim)
 
-        if len(src) == 1:
-            # If the source has a single element, it might be a non-multiplicative unit
-            # and therefore it is treated differently.
-            src_unit, src_value = list(src.items())[0]
-            src_unit = self._units[self.get_name(src_unit)]
+        # Conversion needs to consider if non-multiplicative (AKA offset
+        # units) are involved. Conversion is only possible if src and dst
+        # have at most one offset unit per dimension.
+        src_offset_units = [(u, e) for u, e in src.items()
+                            if not self._units[u].is_multiplicative]
+        dst_offset_units = [(u, e) for u, e in dst.items()
+                            if not self._units[u].is_multiplicative]
 
-            # We only continue if is a ScaleConverter,
-            # if not just exit to use the standard src / dst.
-            # TODO: This will fail and should not degK * meter / nanometer -> degC
-            if not isinstance(src_unit.converter, ScaleConverter):
+        # For offset units we need to check if the conversion is allowed.
+        if src_offset_units or dst_offset_units:
 
-                if len(dst) > 1:
-                    # If the destination has more than one element,
-                    # then the conversion is not possible.
-                    # TODO: This will fail and should not degC -> degK * meter / nanometer
-                    raise DimensionalityError(src, dst, src_dim, dst_dim)
+            # Validate that not more than one offset unit is present
+            if len(src_offset_units) > 1 or len(dst_offset_units) > 1:
+                raise DimensionalityError(
+                    src, dst, src_dim, dst_dim,
+                    extra_msg=' - more than one offset unit.')
 
-                dst_unit, dst_value = list(dst.items())[0]
-                dst_unit = self._units[self.get_name(dst_unit)]
-                if not type(src_unit.converter) is type(dst_unit.converter):
-                    raise DimensionalityError(src, dst, src_dim, dst_dim)
+            # validate that offset unit is not used in multiplicative context
+            if ((len(src_offset_units) == 1 and len(src) > 1)
+                    or (len(dst_offset_units) == 1 and len(dst) > 1)
+                    and not self.autoconvert_offset_to_baseunit):
+                raise DimensionalityError(
+                    src, dst, src_dim, dst_dim,
+                    extra_msg=' - offset unit used in multiplicative context.')
 
-                return dst_unit.converter.from_reference(src_unit.converter.to_reference(value, inplace), inplace)
+            # Validate that order of offset unit is exactly one.
+            if src_offset_units:
+                if src_offset_units[0][1] != 1:
+                    raise DimensionalityError(
+                        src, dst, src_dim, dst_dim,
+                        extra_msg=' - offset units in higher order.')
+            else:
+                if dst_offset_units[0][1] != 1:
+                    raise DimensionalityError(
+                        src, dst, src_dim, dst_dim,
+                        extra_msg=' - offset units in higher order.')
 
+        # Here we convert only the offset quantities. Any remaining scaled
+        # quantities will be converted later.
+
+        # clean src from offset units by converting to reference
+        for u, e in src_offset_units:
+            value = self._units[u].converter.to_reference(value, inplace)
+            src.pop(u)
+
+        # clean dst units from offset units
+        for u, e in dst_offset_units:
+            dst.pop(u)
+
+        # Here src and dst have only multiplicative units left. Thus we can
+        # convert with a factor.
         factor, units = self.get_base_units(src / dst)
 
-        if factor is None:
-            raise DimensionalityError(src, dst, src_dim, dst_dim,
-                                      'Non-multiplicative unit found.')
-
         # factor is type float and if our magnitude is type Decimal then
         # must first convert to Decimal before we can '*' the values
         if isinstance(value, Decimal):
@@ -1037,6 +1103,16 @@ class UnitRegistry(object):
         else:
             value = value * factor
 
+        # Finally convert to offset units specified in destination
+        for u, e in dst_offset_units:
+            value = self._units[u].converter.from_reference(value, inplace)
+            # add back offset units to dst
+            dst[u] = e
+
+        # restore offset conversion of src units
+        for u, e in src_offset_units:
+            src[u] = e
+
         return value
 
     def pi_theorem(self, quantities):
@@ -1069,9 +1145,10 @@ class UnitRegistry(object):
         """Parse a unit to identify prefix, unit name and suffix
         by walking the list of prefix and suffix.
         """
-
-        for suffix, prefix in itertools.product(self._suffixes.keys(), self._prefixes.keys()):
-            if unit_name.startswith(prefix) and unit_name.endswith(suffix):
+        stw = unit_name.startswith
+        edw = unit_name.endswith
+        for suffix, prefix in itertools.product(self._suffixes, self._prefixes):
+            if stw(prefix) and edw(suffix):
                 name = unit_name[len(prefix):]
                 if suffix:
                     name = name[:-len(suffix)]
@@ -1079,30 +1156,33 @@ class UnitRegistry(object):
                         continue
                 if case_sensitive:
                     if name in self._units:
-                        yield (self._prefixes[prefix].name,
-                               self._units[name].name,
+                        yield (self._prefixes[prefix]._name,
+                               self._units[name]._name,
                                self._suffixes[suffix])
                 else:
                     for real_name in self._units_casei.get(name.lower(), ()):
-                        yield (self._prefixes[prefix].name,
-                               self._units[real_name].name,
+                        yield (self._prefixes[prefix]._name,
+                               self._units[real_name]._name,
                                self._suffixes[suffix])
 
-    def parse_units(self, input_string, to_delta=None):
+    def parse_units(self, input_string, as_delta=None):
         """Parse a units expression and returns a UnitContainer with
         the canonical names.
 
         The expression can only contain products, ratios and powers of units.
 
-        :param to_delta: if the expression has multiple units, the parser will
+        :param as_delta: if the expression has multiple units, the parser will
                          interpret non multiplicative units as their `delta_` counterparts.
 
         :raises:
             :class:`pint.UndefinedUnitError` if a unit is not in the registry
             :class:`ValueError` if the expression is invalid.
         """
-        if to_delta is None:
-            to_delta = self.default_to_delta
+        if input_string in self._parse_unit_cache:
+            return self._parse_unit_cache[input_string]
+
+        if as_delta is None:
+            as_delta = self.default_as_delta
 
         if not input_string:
             return UnitsContainer()
@@ -1113,16 +1193,19 @@ class UnitRegistry(object):
 
         ret = UnitsContainer()
         many = len(units) > 1
-        for name, value in units.items():
+        for name in units:
             cname = self.get_name(name)
+            value = units[name]
             if not cname:
                 continue
-            if to_delta and (many or (not many and abs(value) != 1)):
+            if as_delta and (many or (not many and value != 1)):
                 definition = self._units[cname]
                 if not definition.is_multiplicative:
                     cname = 'delta_' + cname
             ret[cname] = value
 
+        self._parse_unit_cache[input_string] = ret
+
         return ret
 
     def parse_expression(self, input_string, case_sensitive=True, **values):
@@ -1141,7 +1224,7 @@ class UnitRegistry(object):
         unknown = set()
         for toknum, tokval, _, _, _ in gen:
             if toknum == NAME:
-                # TODO: Integrate math better, Replace eval
+                # TODO: Integrate math better, Replace eval, make as_delta-aware
                 if tokval == 'pi' or tokval in values:
                     result.append((toknum, tokval))
                     continue
